@@ -1,15 +1,11 @@
-"""
-See Portable Document Format Reference Manual, 1993. ISBN 0-201-62628-4.
+"""Various constants, enums, and flags to aid readability."""
 
-See https://ia802202.us.archive.org/8/items/pdfy-0vt8s-egqFwDl7L2/PDF%20Reference%201.0.pdf
+from enum import Enum, IntFlag, auto, unique
 
-PDF Reference, third edition, Version 1.4, 2001. ISBN 0-201-75839-3.
 
-PDF Reference, sixth edition, Version 1.7, 2006.
-"""
-
-from enum import IntFlag, auto
-from typing import Dict, Tuple
+class StrEnum(str, Enum):  # Once we are on Python 3.11+: enum.StrEnum
+    def __str__(self) -> str:
+        return str(self.value)
 
 
 class Core:
@@ -23,11 +19,12 @@ class Core:
 
 
 class TrailerKeys:
+    SIZE = "/Size"
+    PREV = "/Prev"
     ROOT = "/Root"
     ENCRYPT = "/Encrypt"
-    ID = "/ID"
     INFO = "/Info"
-    SIZE = "/Size"
+    ID = "/ID"
 
 
 class CatalogAttributes:
@@ -39,18 +36,22 @@ class EncryptionDictAttributes:
     """
     Additional encryption dictionary entries for the standard security handler.
 
-    TABLE 3.19, Page 122
+    Table 3.19, Page 122.
+    Table 21 of the 2.0 manual.
     """
 
     R = "/R"  # number, required; revision of the standard security handler
-    O = "/O"  # 32-byte string, required  # noqa
+    O = "/O"  # 32-byte string, required  # noqa: E741
     U = "/U"  # 32-byte string, required
     P = "/P"  # integer flag, required; permitted operations
     ENCRYPT_METADATA = "/EncryptMetadata"  # boolean flag, optional
 
 
 class UserAccessPermissions(IntFlag):
-    """TABLE 3.20 User access permissions."""
+    """
+    Table 3.20 User access permissions.
+    Table 22 of the 2.0 manual.
+    """
 
     R1 = 1
     R2 = 2
@@ -95,9 +96,9 @@ class UserAccessPermissions(IntFlag):
         """Check if the given reserved name defaults to 1 = active."""
         return name not in {"R1", "R2"}
 
-    def to_dict(self) -> Dict[str, bool]:
+    def to_dict(self) -> dict[str, bool]:
         """Convert the given flag value to a corresponding verbose name mapping."""
-        result: Dict[str, bool] = {}
+        result: dict[str, bool] = {}
         for name, flag in UserAccessPermissions.__members__.items():
             if UserAccessPermissions._is_reserved(name):
                 continue
@@ -105,7 +106,7 @@ class UserAccessPermissions(IntFlag):
         return result
 
     @classmethod
-    def from_dict(cls, value: Dict[str, bool]) -> "UserAccessPermissions":
+    def from_dict(cls, value: dict[str, bool]) -> "UserAccessPermissions":
         """Convert the verbose name mapping to the corresponding flag value."""
         value_copy = value.copy()
         result = cls(0)
@@ -127,8 +128,11 @@ class UserAccessPermissions(IntFlag):
         return cls((2**32 - 1) - cls.R1 - cls.R2)
 
 
-class Ressources:
-    """TABLE 3.30 Entries in a resource dictionary."""
+class Resources:
+    """
+    Table 3.30 Entries in a resource dictionary.
+    Table 34 in the 2.0 reference.
+    """
 
     EXT_G_STATE = "/ExtGState"  # dictionary, optional
     COLOR_SPACE = "/ColorSpace"  # dictionary, optional
@@ -141,16 +145,18 @@ class Ressources:
 
 
 class PagesAttributes:
-    """Page Attributes, Table 6.2, Page 52."""
+    """§7.7.3.2 of the 1.7 and 2.0 reference."""
 
     TYPE = "/Type"  # name, required; must be /Pages
-    KIDS = "/Kids"  # array, required; List of indirect references
-    COUNT = "/Count"  # integer, required; the number of all nodes und this node
     PARENT = "/Parent"  # dictionary, required; indirect reference to pages object
+    KIDS = "/Kids"  # array, required; List of indirect references
+    COUNT = "/Count"
+    # integer, required; the number of leaf nodes (page objects)
+    # that are descendants of this node within the page tree
 
 
 class PageAttributes:
-    """TABLE 3.27 Entries in a page object."""
+    """§7.7.3.3 of the 1.7 and 2.0 reference."""
 
     TYPE = "/Type"  # name, required; must be /Page
     PARENT = "/Parent"  # dictionary, required; a pages object
@@ -159,10 +165,10 @@ class PageAttributes:
     )
     RESOURCES = "/Resources"  # dictionary, required if there are any
     MEDIABOX = "/MediaBox"  # rectangle, required; rectangle specifying page size
-    CROPBOX = "/CropBox"  # rectangle, optional; rectangle
-    BLEEDBOX = "/BleedBox"  # rectangle, optional; rectangle
-    TRIMBOX = "/TrimBox"  # rectangle, optional; rectangle
-    ARTBOX = "/ArtBox"  # rectangle, optional; rectangle
+    CROPBOX = "/CropBox"  # rectangle, optional
+    BLEEDBOX = "/BleedBox"  # rectangle, optional
+    TRIMBOX = "/TrimBox"  # rectangle, optional
+    ARTBOX = "/ArtBox"  # rectangle, optional
     BOX_COLOR_INFO = "/BoxColorInfo"  # dictionary, optional
     CONTENTS = "/Contents"  # stream or array, optional
     ROTATE = "/Rotate"  # integer, optional; page rotation in degrees
@@ -178,80 +184,91 @@ class PageAttributes:
     STRUCT_PARENTS = "/StructParents"  # integer, optional
     ID = "/ID"  # byte string, optional
     PZ = "/PZ"  # number, optional
+    SEPARATION_INFO = "/SeparationInfo"  # dictionary, optional
     TABS = "/Tabs"  # name, optional
     TEMPLATE_INSTANTIATED = "/TemplateInstantiated"  # name, optional
     PRES_STEPS = "/PresSteps"  # dictionary, optional
     USER_UNIT = "/UserUnit"  # number, optional
     VP = "/VP"  # dictionary, optional
+    AF = "/AF"  # array of dictionaries, optional
+    OUTPUT_INTENTS = "/OutputIntents"  # array, optional
+    D_PART = "/DPart"  # dictionary, required, if this page is within the range of a DPart, not permitted otherwise
 
 
 class FileSpecificationDictionaryEntries:
-    """TABLE 3.41 Entries in a file specification dictionary."""
+    """Table 3.41 Entries in a file specification dictionary."""
 
     Type = "/Type"
     FS = "/FS"  # The name of the file system to be used to interpret this file specification
-    F = "/F"  # A file specification string of the form described in Section 3.10.1
-    UF = "/UF"  # A unicode string of the file as described in Section 3.10.1
+    F = "/F"  # A file specification string of the form described in §3.10.1
+    UF = "/UF"  # A Unicode string of the file as described in §3.10.1
     DOS = "/DOS"
     Mac = "/Mac"
     Unix = "/Unix"
     ID = "/ID"
     V = "/V"
-    EF = "/EF"  # dictionary, containing a subset of the keys F , UF , DOS , Mac , and Unix
+    EF = "/EF"  # dictionary, containing a subset of the keys F, UF, DOS, Mac, and Unix
     RF = "/RF"  # dictionary, containing arrays of /EmbeddedFile
     DESC = "/Desc"  # description of the file
     Cl = "/Cl"
 
 
 class StreamAttributes:
-    """Table 4.2."""
+    """
+    Table 4.2.
+    Table 5 in the 2.0 reference.
+    """
 
     LENGTH = "/Length"  # integer, required
     FILTER = "/Filter"  # name or array of names, optional
     DECODE_PARMS = "/DecodeParms"  # variable, optional -- 'decodeParams is wrong
 
 
-class FilterTypes:
-    """
-    Table 4.3 of the 1.4 Manual.
-
-    Page 354 of the 1.7 Manual
-    """
+@unique
+class FilterTypes(StrEnum):
+    """§7.4 of the 1.7 and 2.0 references."""
 
     ASCII_HEX_DECODE = "/ASCIIHexDecode"  # abbreviation: AHx
     ASCII_85_DECODE = "/ASCII85Decode"  # abbreviation: A85
     LZW_DECODE = "/LZWDecode"  # abbreviation: LZW
-    FLATE_DECODE = "/FlateDecode"  # abbreviation: Fl, PDF 1.2
+    FLATE_DECODE = "/FlateDecode"  # abbreviation: Fl
     RUN_LENGTH_DECODE = "/RunLengthDecode"  # abbreviation: RL
     CCITT_FAX_DECODE = "/CCITTFaxDecode"  # abbreviation: CCF
     DCT_DECODE = "/DCTDecode"  # abbreviation: DCT
     JPX_DECODE = "/JPXDecode"
+    JBIG2_DECODE = "/JBIG2Decode"
 
 
 class FilterTypeAbbreviations:
-    """Table 4.44 of the 1.7 Manual (page 353ff)."""
+    """§8.9.7 of the 1.7 and 2.0 references."""
 
     AHx = "/AHx"
     A85 = "/A85"
     LZW = "/LZW"
-    FL = "/Fl"  # FlateDecode
+    FL = "/Fl"
     RL = "/RL"
     CCF = "/CCF"
     DCT = "/DCT"
 
 
 class LzwFilterParameters:
-    """Table 4.4."""
+    """
+    Table 4.4.
+    Table 8 in the 2.0 reference.
+    """
 
     PREDICTOR = "/Predictor"  # integer
-    COLUMNS = "/Columns"  # integer
     COLORS = "/Colors"  # integer
     BITS_PER_COMPONENT = "/BitsPerComponent"  # integer
+    COLUMNS = "/Columns"  # integer
     EARLY_CHANGE = "/EarlyChange"  # integer
 
 
 class CcittFaxDecodeParameters:
-    """Table 4.5."""
+    """
+    Table 4.5.
+    Table 11 in the 2.0 reference.
+    """
 
     K = "/K"  # integer
     END_OF_LINE = "/EndOfLine"  # boolean
@@ -264,7 +281,7 @@ class CcittFaxDecodeParameters:
 
 
 class ImageAttributes:
-    """Table 4.39 Pdf Reference 1.7 page 340+"""
+    """§11.6.5 of the 1.7 and 2.0 references."""
 
     TYPE = "/Type"  # name, required; must be /XObject
     SUBTYPE = "/Subtype"  # name, required; must be /Image
@@ -299,23 +316,24 @@ class TypArguments:
 class TypFitArguments:
     """Table 8.2 of the PDF 1.7 reference."""
 
-    FIT = "/Fit"
-    FIT_V = "/FitV"
-    FIT_BV = "/FitBV"
-    FIT_B = "/FitB"
-    FIT_H = "/FitH"
-    FIT_BH = "/FitBH"
-    FIT_R = "/FitR"
     XYZ = "/XYZ"
+    FIT = "/Fit"
+    FIT_H = "/FitH"
+    FIT_V = "/FitV"
+    FIT_R = "/FitR"
+    FIT_B = "/FitB"
+    FIT_BH = "/FitBH"
+    FIT_BV = "/FitBV"
 
 
 class GoToActionArguments:
     S = "/S"  # name, required: type of action
-    D = "/D"  # name / byte string /array, required: Destination to jump to
+    D = "/D"  # name, byte string, or array, required: destination to jump to
+    SD = "/SD"  # array, optional: structure destination to jump to
 
 
 class AnnotationDictionaryAttributes:
-    """TABLE 8.15 Entries common to all annotation dictionaries."""
+    """Table 8.15 Entries common to all annotation dictionaries."""
 
     Type = "/Type"
     Subtype = "/Subtype"
@@ -346,7 +364,12 @@ class InteractiveFormDictEntries:
 
 
 class FieldDictionaryAttributes:
-    """TABLE 8.69 Entries common to all field dictionaries (PDF 1.7 reference)."""
+    """
+    Entries common to all field dictionaries (Table 8.69 PDF 1.7 reference)
+    (*very partially documented here*).
+
+    FFBits provides the constants used for `/Ff` from Table 8.70/8.75/8.77/8.79
+    """
 
     FT = "/FT"  # name, required for terminal fields
     Parent = "/Parent"  # dictionary, required for children
@@ -358,39 +381,68 @@ class FieldDictionaryAttributes:
     V = "/V"  # text string or array, optional
     DV = "/DV"  # text string, optional
     AA = "/AA"  # dictionary, optional
-    Opt = "/Opt"
+    Opt = "/Opt"  # array, optional
+
+    class FfBits(IntFlag):
+        """
+        Ease building /Ff flags
+        Some entries may be specific to:
+
+        * Text (Tx) (Table 8.75 PDF 1.7 reference)
+        * Buttons (Btn) (Table 8.77 PDF 1.7 reference)
+        * Choice (Ch) (Table 8.79 PDF 1.7 reference)
+        """
 
-    class FfBits:
         ReadOnly = 1 << 0
+        """common to Tx/Btn/Ch in Table 8.70"""
         Required = 1 << 1
+        """common to Tx/Btn/Ch in Table 8.70"""
         NoExport = 1 << 2
-        Multiline = 1 << 12  # Tx Table 8.77
-        Password = 1 << 13  # Tx
+        """common to Tx/Btn/Ch in Table 8.70"""
 
-        NoToggleToOff = 1 << 14  # Btn table 8.75
-        Radio = 1 << 15  # Btn
-        Pushbutton = 1 << 16  # Btn
+        Multiline = 1 << 12
+        """Tx"""
+        Password = 1 << 13
+        """Tx"""
 
-        Combo = 1 << 17  # Ch table 8.79
-        Edit = 1 << 18  # Ch
-        Sort = 1 << 19  # Ch
+        NoToggleToOff = 1 << 14
+        """Btn"""
+        Radio = 1 << 15
+        """Btn"""
+        Pushbutton = 1 << 16
+        """Btn"""
 
-        FileSelect = 1 << 20  # Tx
+        Combo = 1 << 17
+        """Ch"""
+        Edit = 1 << 18
+        """Ch"""
+        Sort = 1 << 19
+        """Ch"""
 
-        MultiSelect = 1 << 21  # Ch
+        FileSelect = 1 << 20
+        """Tx"""
 
-        DoNotSpellCheck = 1 << 22  # Tx / Ch
-        DoNotScroll = 1 << 23  # Tx
-        Comb = 1 << 24  # Tx
+        MultiSelect = 1 << 21
+        """Tx"""
 
-        RadiosInUnison = 1 << 25  # Btn
+        DoNotSpellCheck = 1 << 22
+        """Tx/Ch"""
+        DoNotScroll = 1 << 23
+        """Tx"""
+        Comb = 1 << 24
+        """Tx"""
 
-        RichText = 1 << 25  # Tx
+        RadiosInUnison = 1 << 25
+        """Btn"""
 
-        CommitOnSelChange = 1 << 26  # Ch
+        RichText = 1 << 25
+        """Tx"""
+
+        CommitOnSelChange = 1 << 26
+        """Ch"""
 
     @classmethod
-    def attributes(cls) -> Tuple[str, ...]:
+    def attributes(cls) -> tuple[str, ...]:
         """
         Get a tuple of all the attributes present in a Field Dictionary.
 
@@ -401,6 +453,7 @@ class FieldDictionaryAttributes:
 
         Returns:
             A tuple containing all the attribute constants.
+
         """
         return (
             cls.TM,
@@ -416,7 +469,7 @@ class FieldDictionaryAttributes:
         )
 
     @classmethod
-    def attributes_dict(cls) -> Dict[str, str]:
+    def attributes_dict(cls) -> dict[str, str]:
         """
         Get a dictionary of attribute keys and their human-readable names.
 
@@ -428,6 +481,7 @@ class FieldDictionaryAttributes:
 
         Returns:
             A dictionary containing attribute keys and their names.
+
         """
         return {
             cls.FT: "Field Type",
@@ -442,12 +496,12 @@ class FieldDictionaryAttributes:
 
 
 class CheckboxRadioButtonAttributes:
-    """TABLE 8.76 Field flags common to all field types."""
+    """Table 8.76 Field flags common to all field types."""
 
     Opt = "/Opt"  # Options, Optional
 
     @classmethod
-    def attributes(cls) -> Tuple[str, ...]:
+    def attributes(cls) -> tuple[str, ...]:
         """
         Get a tuple of all the attributes present in a Field Dictionary.
 
@@ -458,11 +512,12 @@ class CheckboxRadioButtonAttributes:
 
         Returns:
             A tuple containing all the attribute constants.
+
         """
         return (cls.Opt,)
 
     @classmethod
-    def attributes_dict(cls) -> Dict[str, str]:
+    def attributes_dict(cls) -> dict[str, str]:
         """
         Get a dictionary of attribute keys and their human-readable names.
 
@@ -474,6 +529,7 @@ class CheckboxRadioButtonAttributes:
 
         Returns:
             A dictionary containing attribute keys and their names.
+
         """
         return {
             cls.Opt: "Options",
@@ -481,7 +537,7 @@ class CheckboxRadioButtonAttributes:
 
 
 class FieldFlag(IntFlag):
-    """TABLE 8.70 Field flags common to all field types."""
+    """Table 8.70 Field flags common to all field types."""
 
     READ_ONLY = 1
     REQUIRED = 2
@@ -489,7 +545,7 @@ class FieldFlag(IntFlag):
 
 
 class DocumentInformationAttributes:
-    """TABLE 10.2 Entries in the document information dictionary."""
+    """Table 10.2 Entries in the document information dictionary."""
 
     TITLE = "/Title"  # text string, optional
     AUTHOR = "/Author"  # text string, optional
@@ -503,12 +559,17 @@ class DocumentInformationAttributes:
 
 
 class PageLayouts:
-    """Page 84, PDF 1.4 reference."""
+    """
+    Page 84, PDF 1.4 reference.
+    Page 115, PDF 2.0 reference.
+    """
 
     SINGLE_PAGE = "/SinglePage"
     ONE_COLUMN = "/OneColumn"
     TWO_COLUMN_LEFT = "/TwoColumnLeft"
     TWO_COLUMN_RIGHT = "/TwoColumnRight"
+    TWO_PAGE_LEFT = "/TwoPageLeft"  # (PDF 1.5)
+    TWO_PAGE_RIGHT = "/TwoPageRight"  # (PDF 1.5)
 
 
 class GraphicsStateParameters:
@@ -544,10 +605,11 @@ class GraphicsStateParameters:
 
 
 class CatalogDictionary:
-    """Table 3.25 in the 1.7 reference."""
+    """§7.7.2 of the 1.7 and 2.0 references."""
 
     TYPE = "/Type"  # name, required; must be /Catalog
     VERSION = "/Version"  # name
+    EXTENSIONS = "/Extensions"  # dictionary, optional; ISO 32000-1
     PAGES = "/Pages"  # dictionary, required
     PAGE_LABELS = "/PageLabels"  # number tree, optional
     NAMES = "/Names"  # dictionary, optional
@@ -574,6 +636,9 @@ class CatalogDictionary:
     REQUIREMENTS = "/Requirements"  # array, optional
     COLLECTION = "/Collection"  # dictionary, optional
     NEEDS_RENDERING = "/NeedsRendering"  # boolean, optional
+    DSS = "/DSS"  # dictionary, optional
+    AF = "/AF"  # array of dictionaries, optional
+    D_PART_ROOT = "/DPartRoot"  # dictionary, optional
 
 
 class OutlineFontFlag(IntFlag):
@@ -584,17 +649,20 @@ class OutlineFontFlag(IntFlag):
 
 
 class PageLabelStyle:
-    """Table 8.10 in the 1.7 reference."""
+    """
+    Table 8.10 in the 1.7 reference.
+    Table 161 in the 2.0 reference.
+    """
 
-    DECIMAL = "/D"  # Decimal arabics
-    LOWERCASE_ROMAN = "/r"  # Lowercase roman numbers
-    UPPERCASE_ROMAN = "/R"  # Uppercase roman numbers
-    LOWERCASE_LETTER = "/a"  # Lowercase letters
+    DECIMAL = "/D"  # Decimal Arabic numerals
+    UPPERCASE_ROMAN = "/R"  # Uppercase Roman numerals
+    LOWERCASE_ROMAN = "/r"  # Lowercase Roman numerals
     UPPERCASE_LETTER = "/A"  # Uppercase letters
+    LOWERCASE_LETTER = "/a"  # Lowercase letters
 
 
 class AnnotationFlag(IntFlag):
-    """See 12.5.3 "Annotation Flags"."""
+    """See §12.5.3 "Annotation Flags"."""
 
     INVISIBLE = 1
     HIDDEN = 2
@@ -619,17 +687,18 @@ PDF_KEYS = (
     DocumentInformationAttributes,
     EncryptionDictAttributes,
     FieldDictionaryAttributes,
+    FileSpecificationDictionaryEntries,
     FilterTypeAbbreviations,
     FilterTypes,
     GoToActionArguments,
     GraphicsStateParameters,
     ImageAttributes,
-    FileSpecificationDictionaryEntries,
+    InteractiveFormDictEntries,
     LzwFilterParameters,
     PageAttributes,
     PageLayouts,
     PagesAttributes,
-    Ressources,
+    Resources,
     StreamAttributes,
     TrailerKeys,
     TypArguments,
@@ -644,3 +713,70 @@ class ImageType(IntFlag):
     DRAWING_IMAGES = auto()
     ALL = XOBJECT_IMAGES | INLINE_IMAGES | DRAWING_IMAGES
     IMAGES = ALL  # for consistency with ObjectDeletionFlag
+
+
+_INLINE_IMAGE_VALUE_MAPPING = {
+    "/G": "/DeviceGray",
+    "/RGB": "/DeviceRGB",
+    "/CMYK": "/DeviceCMYK",
+    "/I": "/Indexed",
+    "/AHx": "/ASCIIHexDecode",
+    "/A85": "/ASCII85Decode",
+    "/LZW": "/LZWDecode",
+    "/Fl": "/FlateDecode",
+    "/RL": "/RunLengthDecode",
+    "/CCF": "/CCITTFaxDecode",
+    "/DCT": "/DCTDecode",
+    "/DeviceGray": "/DeviceGray",
+    "/DeviceRGB": "/DeviceRGB",
+    "/DeviceCMYK": "/DeviceCMYK",
+    "/Indexed": "/Indexed",
+    "/ASCIIHexDecode": "/ASCIIHexDecode",
+    "/ASCII85Decode": "/ASCII85Decode",
+    "/LZWDecode": "/LZWDecode",
+    "/FlateDecode": "/FlateDecode",
+    "/RunLengthDecode": "/RunLengthDecode",
+    "/CCITTFaxDecode": "/CCITTFaxDecode",
+    "/DCTDecode": "/DCTDecode",
+    "/RelativeColorimetric": "/RelativeColorimetric",
+}
+
+_INLINE_IMAGE_KEY_MAPPING = {
+    "/BPC": "/BitsPerComponent",
+    "/CS": "/ColorSpace",
+    "/D": "/Decode",
+    "/DP": "/DecodeParms",
+    "/F": "/Filter",
+    "/H": "/Height",
+    "/W": "/Width",
+    "/I": "/Interpolate",
+    "/Intent": "/Intent",
+    "/IM": "/ImageMask",
+    "/BitsPerComponent": "/BitsPerComponent",
+    "/ColorSpace": "/ColorSpace",
+    "/Decode": "/Decode",
+    "/DecodeParms": "/DecodeParms",
+    "/Filter": "/Filter",
+    "/Height": "/Height",
+    "/Width": "/Width",
+    "/Interpolate": "/Interpolate",
+    "/ImageMask": "/ImageMask",
+}
+
+
+class AFRelationship:
+    """
+    Associated file relationship types, defining the relationship between
+    the PDF component and the associated file.
+
+    Defined in table 43 of the PDF 2.0 reference.
+    """
+
+    SOURCE = "/Source"  # Original content source
+    DATA = "/Data"  # Base data for visual presentation
+    ALTERNATIVE = "/Alternative"  # Alternative content representation
+    SUPPLEMENT = "/Supplement"  # Supplemental representation of original source/data
+    ENCRYPTED_PAYLOAD = "/EncryptedPayload"  # Encrypted payload document
+    FORM_DATA = "/FormData"  # Data associated with AcroForm of this PDF
+    SCHEMA = "/Schema"  # Schema definition for associated object
+    UNSPECIFIED = "/Unspecified"  # Not known or cannot be described with values
